##
# File:    CcdcGeomAnal.py
# Author:  J. Westbrook
# Date:    17-Dec-2020
# Version: 0.001
#
# Updated:
#
##
# pylint: disable=exec-used
"""
Utilities for chemical component geometrical analyis using the local CCDC Python API
"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
import sys

from ccdc.io import EntryReader
from ccdc import conformer

logger = logging.getLogger(__name__)


class CcdcGeomAnal(object):
    def __init__(self, verbose=True, log=sys.stderr):
        self.__lfh = log
        self.__verbose = verbose
        self.__engine = conformer.GeometryAnalyser()

    def settings(self):
        return self.__engine.settings.summary()

    def featureSettings(self, featureType, **kw):
        """
        Features specific settings -

            featureType = 'bond', 'angle', 'torsion', 'ring'

        Features and values:

            few_hits_threshold: Threshold below which a distribution is considered to have too few hits.

            local_density_threshold: Local density threshold used to classify torsions and rings as unusual.

                                    Note that the local density is irrelevant for bonds and angles.

            local_density_tolerance: The local density tolerance.

            min_obs_exact: Minimum acceptable size of an exact distribution.
                            If there is no distribution containing at least this number of observations the
                            geometry analyser will perform a generalised search according to the criteria specified by other settings.

            min_obs_generalised: Minimum number of observations that the geometry analyser should try to find.
                                If this is 0 then generalised searches will never be performed.
                                Similarly, if generalisation has been turned off this setting will not have an effect.

            min_relevance:  Relevance criterion for a generalised hit to be accepted.

                            The geometry analyser determines how similar a fragment is to the query by calculating
                            a relevance value. The min_relevance setting tells the geometry analyser to accept,
                            in a generalised search, only fragments whose relevance is equal to or greater than this threshold.

            zscore_threshold:  Z-score thresh

            classification_measure: Z-score : Local density, Nearest observation, Mean distance

            classification_measure_threshold:  float


        """
        fTypes = ["bond", "angle", "torsion", "ring"]
        fNames = ["few_hits_threshold", "local_density_threshold", "local_density_tolerance", "min_obs_exact", "min_obs_generalised", "min_relevance", "zscore_threshold"]

        if featureType not in fTypes:
            return False

        for k in kw:
            if k in fNames:
                cm = "e.settings.%s.%s=%s" % (featureType, k, str(kw[k]))
                logger.info("Settings command: %s", cm)
                self.__doExec(self.__engine, cm)

        return True

    def globalSettings(self, **kw):
        """
        Global setting keywords and values:

            rfactor_filter: 0.05,0.075,0.1,any
            generalisation True: False
            organometallic_filter : all,metalorganics_only, organics_only
            solvent_filter: include_solvent,exclude_solvent,only_solvent
            heaviest_element:  <element symbol> (case sensistive)
        """
        fNames = ["rfactor_filter", "generalisation"]
        sNames = ["organometallic_filter", "solvent_filter", "heaviest_element"]

        for k in kw:
            if k in fNames:
                cm = "e.settings.%s=%s" % (k, str(kw[k]))
                logger.info("Settings command: %s", cm)
                self.__doExec(self.__engine, cm)
            if k in sNames:
                cm = "e.settings.%s='%s'" % (k, str(kw[k]))
                logger.info("Settings command: %s", cm)
                self.__doExec(self.__engine, cm)
        return True

    def anal(self, queryTargetPath, normalizeFlag=False):
        """Perform geometrical analysis against the CCDC data source-"""
        retD = {}
        targetStructures = EntryReader(queryTargetPath)

        for e in targetStructures:
            mol = e.molecule
            if normalizeFlag:
                mol.assign_bond_types(which="unknown")
                mol.standardise_aromatic_bonds()
                mol.standardise_delocalised_bonds()
            #
            logger.info("begin analysis - for %s", queryTargetPath)
            gam = self.__engine.analyse_molecule(mol)
            bondOutliers = len([b for b in gam.analysed_bonds if b.unusual and b.enough_hits])
            angleOutliers = len([a for a in gam.analysed_angles if a.unusual and a.enough_hits])
            torsionOutliers = len([t for t in gam.analysed_torsions if t.unusual and t.enough_hits])
            ringOutliers = len([r for r in gam.analysed_rings if r.unusual and r.enough_hits])

            bL = self.__getBondAnalysis(gam)
            aL = self.__getAngleAnalysis(gam)
            tL = self.__getTorsionAnalysis(gam)
            rL = self.__getRingAnalysis(gam)
            retD = {
                "bond_outliers": bondOutliers,
                "angle_outliers": angleOutliers,
                "torsion_outliers": torsionOutliers,
                "ring_outliers": ringOutliers,
                "bond_list": bL,
                "angle_list": aL,
                "torsion_list": tL,
                "ring_list": rL,
            }
        return retD

    def __extractAnalFeatures(self, feature):
        rD = {
            "atom_labels": feature.atom_labels,
            "d_min": feature.d_min,
            "value": feature.value,
            "mean": feature.mean,
            "standard_deviation": feature.standard_deviation,
            "z_score": feature.z_score,
            "type": feature.type,
            "median": feature.median,
            "lower_quartile": feature.lower_quartile,
            "upper_quartile": feature.upper_quartile,
            "minimum": feature.minimum,
            "maximum": feature.maximum,
            "nhits": feature.nhits,
            "unusual": feature.unusual,
            "generalised": feature.generalised,
            "few_hits": feature.few_hits,
            "enough_hits": feature.enough_hits,
            "local_density": feature.local_density,
        }
        return rD

    def __getBondAnalysis(self, mol):
        rL = []
        for feature in mol.analysed_bonds:
            rL.append(self.__extractAnalFeatures(feature))
        return rL

    def __getAngleAnalysis(self, mol):
        rL = []
        for feature in mol.analysed_angles:
            rL.append(self.__extractAnalFeatures(feature))
        return rL

    def __getTorsionAnalysis(self, mol):
        rL = []
        for feature in mol.analysed_torsions:
            rL.append(self.__extractAnalFeatures(feature))
        return rL

    def __getRingAnalysis(self, mol):
        rL = []
        for feature in mol.analysed_rings:
            rL.append(self.__extractAnalFeatures(feature))
        return rL

    def __doExec(self, e, cmd):
        _ = e
        exec(cmd, globals(), locals())



##
# File: CcdcSearchExec.py
# Date: 15-Dec-2020  jdw
#
#  Execution wrapper  --  for CCDC search operations (wraps up the py37 environment)
#
#  Updates:
#   15-Jan-2021 jdw add option to export search hit list.
#
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import argparse
import logging
import os
import sys

from rcsb.utils.io.MarshalUtil import MarshalUtil

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


def main():
    parser = argparse.ArgumentParser()
    #
    parser.add_argument("--mol_list_path", default=None, help="Molecule file list path")
    parser.add_argument("--result_path", default=None, help="Molecule file list path")
    parser.add_argument("--search_type", default=None, help="Search type (similarity|substructure)")
    parser.add_argument("--start_record", default=None, help="Starting record")
    parser.add_argument("--end_record", default=None, help="End record")
    parser.add_argument("--csdhome", default=None, help="Path to the CSD release (path to CSD_202x)")
    parser.add_argument("--python_lib_path", default=None, help="Path to Python library")
    parser.add_argument("--python_version", default=None, help="Python library version (default: 3.7)")
    parser.add_argument("--hit_list_path", default=None, help="Path to list of molecule identifers with search results")
    #
    args = parser.parse_args()
    #
    try:
        pyLib = args.python_lib_path if args.python_lib_path else os.path.join(os.environ["PYENV_ROOT"], "versions", "3.7.9", "lib")
        pyVer = args.python_version if args.python_version else "3.7"

        csdHome = args.csdhome
        molFilePath = args.mol_list_path
        resultPath = args.result_path
        searchType = args.search_type
        startRecord = args.start_record
        endRecord = args.end_record
        hitListPath = args.hit_list_path
    except Exception as e:
        logger.exception("Argument processing problem %s", str(e))
        parser.print_help(sys.stderr)
        exit(1)
    #
    try:
        os.environ["CSDHOME"] = csdHome
        os.environ["LD_LIBRARY_PATH"] = "%s:%s/python%s/site-packages/ccdc/_lib:$LD_LIBRARY_PATH" % (pyLib, pyLib, pyVer)
        os.environ["DYLD_LIBRARY_PATH"] = "%s/python%s/site-packages/ccdc/_lib" % (pyLib, pyVer)
        os.environ["DYLD_FRAMEWORK_PATH"] = "%s/python%s/site-packages/ccdc/_lib" % (pyLib, pyVer)

        logger.info("Using CSDHOME %s", os.environ["CSDHOME"])
        logger.info("Using DYLD_LIBRARY_PATH %s", os.environ["DYLD_LIBRARY_PATH"])
        logger.info("Using DYLD_FRAMEWORK_PATH %s", os.environ["DYLD_FRAMEWORK_PATH"])

        from rcsb.utils.ccdc.CcdcSearch import CcdcSearch  # pylint: disable=import-outside-toplevel

        ccdcS = CcdcSearch(verbose=True)
        pL = ccdcS.getList(molFilePath, startRecord=startRecord, endRecord=endRecord)
        logger.info("Search file %s record length %r", molFilePath, len(pL) if pL else [])
        #
        hitL = []
        for ii, queryTargetPath in enumerate(pL, 1):
            _, fn = os.path.split(queryTargetPath)
            queryTargetId, _ = os.path.splitext(fn)
            #
            logger.info("(%d/%d) Start search for %r %r", ii, len(pL), queryTargetId, queryTargetPath)
            numHits = ccdcS.search(queryTargetId, queryTargetPath, resultPath, searchType=searchType)
            if numHits:
                hitL.append(queryTargetId)
        logger.info("%d searches completed - matched %d", len(pL), len(hitL))
        if hitListPath:
            mU = MarshalUtil()
            ok = mU.doExport(hitListPath, hitL, fmt="list")
            logger.info("Wrote hit list (%r) to %s", ok, hitListPath)
    except Exception as e:
        logger.exception("Failing with %s", str(e))

    # ----------------------- - ----------------------- - ----------------------- - ----------------------- - ----------------------- -


if __name__ == "__main__":
    main()



##
#
# File:    CcdcSearch.py
# Author:  J. Westbrook
# Date:    28-July-2017
# Version: 0.001
#
# Updated:
#   21-Jun-2016   jdw  export additional metadata with each matching structure  -
#   22-Jun-2016   jdw  refactor with general index class CcdcMatchIndex -
#   28-Jul-2017   jdw  Generalize to CcdcSearch.py
#   28-Jul-2017   jdw  remove parentId ---
#
##
"""
Chemical component search against the local CCDC using Python API

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

# pylint: disable=not-context-manager

import copy
import logging
import time
import os

from ccdc.io import EntryReader, MoleculeWriter, csd_version, csd_directory
from ccdc.search import SimilaritySearch, TextNumericSearch, MoleculeSubstructure, SubstructureSearch, SMARTSSubstructure

from rcsb.utils.io.IndexUtils import CcdcMatchIndex, CcdcMatchIndexInst
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class CcdcSearch(object):
    def __init__(self, verbose=True, similarityThreshold=0.95, rValueMaxPercent=10.0):
        self.__verbose = verbose
        self.__similarityThreshold = similarityThreshold
        self.__rValueMaxPercent = rValueMaxPercent

    def getList(self, listPath, startRecord=None, endRecord=None):
        rL = []
        try:
            mU = MarshalUtil()
            logger.debug("Reading path list %r", listPath)
            pL = mU.doImport(listPath, fmt="list")
            logger.debug("Reading path list %r (%r)", listPath, len(pL) if pL else None)
            logger.debug("path list %r", pL)
            if startRecord and endRecord:
                rL = pL[startRecord - 1 : endRecord]
            else:
                # take the full list
                rL = pL
        except Exception as e:
            logger.exception("Failing for %r with %s", listPath, str(e))
        #
        return rL

    def search(self, queryTargetId, queryTargetPath, resultPath, normalizeFlag=True, maxHits=50, searchType="similarity", suppressMetals=False):
        """Search the CCDC database for similar or substructure matches for the input query molecule.

        Args:
            queryTargetId (str): query identifier
            queryTargetPath (str): path to the query molfile (mol, sdf, mol2)
            resultPath (str): output path to match results
            normalizeFlag (bool, optional): do standard perceptions on matching molecules. Defaults to True.
            maxHits (int, optional): maximum number of matches to return. Defaults to 50.
            searchType (str, optional): search mode (substructure, similarity). Defaults to "similarity".
            suppressMetals (bool, optional): filter structures containing metals. Defaults to False.

        Returns:
            (int): number of matches
        """

        mU = MarshalUtil()
        logger.info("Start search for target %s path %s result path %s", queryTargetId, queryTargetPath, resultPath)
        #
        summaryList = []
        #
        targetDirPath = os.path.dirname(queryTargetPath)
        cifTargetPath = os.path.join(targetDirPath, queryTargetId + ".cif")

        #
        targetStructures = EntryReader(queryTargetPath)
        dirPath = os.path.join(resultPath, queryTargetId)
        numHits = 0
        for ii, e in enumerate(targetStructures, 1):
            numHits = 0
            startTime = time.time()
            targetMol = e.molecule
            if normalizeFlag:
                targetMol.assign_bond_types(which="unknown")
                targetMol.standardise_aromatic_bonds()
                targetMol.standardise_delocalised_bonds()
            #
            logger.info("(%d) begin %s search - query id %s", ii, searchType, queryTargetId)
            if searchType == "similarity":
                hits = self.__similaritySearch(targetMol, suppressMetals=suppressMetals)
            elif searchType == "substructure":
                hits = self.__moleculeSubstructureSearch(targetMol, suppressMetals=suppressMetals)
            else:
                hits = []
            logger.info("(%d) completed search query id %s in %.3f seconds", ii, queryTargetId, time.time() - startTime)

            if hits:
                numHits += len(hits)
                logger.info("(%d) search for %s matched %d: %r", ii, queryTargetId, numHits, [targetHit.identifier for targetHit in hits])

                #
                for targetHit in hits[:maxHits]:
                    #
                    hI = CcdcMatchIndexInst()
                    hI.setCsdVersion(csd_version())
                    hI.setCsdDirectory(csd_directory())
                    hI.setTargetId(queryTargetId)
                    hI.setTargetPath(queryTargetPath)
                    if mU.exists(cifTargetPath):
                        hI.setTargetCcPath(cifTargetPath)
                    hI.setIdentifier(targetHit.identifier)
                    hI.setMatchType(searchType)
                    try:
                        hI.setRFactor(targetHit.entry.r_factor)
                        hI.setChemicalName(targetHit.entry.chemical_name)
                        hI.setTemperature(targetHit.entry.temperature)
                        hI.setRadiationSource(targetHit.entry.radiation_source)
                        hI.setHasDisorder("N")
                        cit = targetHit.entry.publication
                        if cit.doi is not None:
                            hI.setCitationDOI(cit.doi)
                        if searchType == "similarity":
                            hI.setSimilarityScore(targetHit.similarity)
                        elif searchType == "substructure":
                            hI.setMatchedAtomLength(len(targetHit.match_atoms()))
                    except Exception as e:
                        logger.exception("Failing with %s", str(e))
                        #
                    #
                    mU.mkdir(dirPath)
                    mol2L = []
                    if searchType == "substructure":
                        for jj, mc in enumerate(targetHit.match_components(), 1):
                            fp = os.path.join(dirPath, queryTargetId + "_" + targetHit.identifier + "_%03d" % jj + ".mol2")
                            mol2L.append(fp)
                            with MoleculeWriter(fp) as ofh:
                                ofh.write(mc)
                            # Replace the title line
                            with open(fp) as fin:
                                lines = fin.readlines()
                            lines[1] = lines[1].replace("00", targetHit.identifier)
                            #
                            with open(fp, "w") as fout:
                                fout.write("".join(lines))
                            #
                            fp = os.path.join(dirPath, queryTargetId + "_" + targetHit.identifier + "_%03d" % jj + ".sdf")
                            with MoleculeWriter(fp) as ofh:
                                ofh.write(mc)

                            # Replace the title line
                            with open(fp) as fin:
                                lines = fin.readlines()
                            lines[0] = lines[0].replace("00", targetHit.identifier)
                            #
                            with open(fp, "w") as fout:
                                fout.write("".join(lines))
                        #
                        #  Check for multiple generated result files -
                        #
                        for jj, fp in enumerate(mol2L, 1):
                            logger.debug("(%d) adding component fp %s", jj, fp)
                            hI.setMatchNumber(jj)
                            hI.setMol2Path(fp)
                            tt = fp[:-4] + "sdf"
                            hI.setMolPath(tt)
                            summaryList.append(copy.deepcopy(hI.get()))
                            #
                    else:
                        hI.setMatchNumber(1)
                        summaryList.append(copy.deepcopy(hI.get()))
            else:
                logger.info("(%d) search for %s returns no matches", ii, targetMol.identifier)
                hits = None
        #
        if numHits > 0:
            mU.mkdir(dirPath)
            fp = os.path.join(dirPath, queryTargetId + "-index.json")
            cmI = CcdcMatchIndex(indexFilePath=fp, verbose=self.__verbose)
            cmI.load(summaryList)
            cmI.writeIndex()

        return numHits

    def searchSmarts(self, queryTargetId, smarts, resultPath, maxHits=50, suppressMetals=False):
        """Search the CCDC database for substructure matches for the input SMARTS pattern.

        Args:
            queryTargetId (str): query identifier
            smarts (srt): smarts search pattern (NON STEREO)
            resultPath (str): output path to match results
            maxHits (int, optional): maximum number of matches to return. Defaults to 50.
            suppressMetals (bool, optional): filter structures containing metals. Defaults to False.

        Returns:
            (int): number of matches
        """

        mU = MarshalUtil()
        logger.info("Start smarts search for target %s result path %s", queryTargetId, resultPath)
        #
        ii = 1
        searchType = "substructure"
        summaryList = []
        dirPath = os.path.join(resultPath, queryTargetId)
        numHits = 0
        startTime = time.time()
        logger.info("(%d) begin %s search - query id %s", ii, searchType, queryTargetId)

        if searchType == "substructure":
            hits = self.__smartsSubstructureSearch(smarts, suppressMetals=suppressMetals)
        else:
            hits = []
        logger.info("(%d) completed search query id %s in %.3f seconds", ii, queryTargetId, time.time() - startTime)

        if hits:
            numHits += len(hits)
            logger.info("(%d) search for %s matched %d: %r", ii, queryTargetId, numHits, [targetHit.identifier for targetHit in hits])

            #
            for targetHit in hits[:maxHits]:
                #
                hI = CcdcMatchIndexInst()
                hI.setTargetId(queryTargetId)
                hI.setIdentifier(targetHit.identifier)
                hI.setMatchType(searchType)
                try:
                    hI.setRFactor(targetHit.entry.r_factor)
                    hI.setChemicalName(targetHit.entry.chemical_name)
                    hI.setTemperature(targetHit.entry.temperature)
                    hI.setRadiationSource(targetHit.entry.radiation_source)
                    hI.setHasDisorder("N")
                    cit = targetHit.entry.publication
                    if cit.doi is not None:
                        hI.setCitationDOI(cit.doi)
                    if searchType == "similarity":
                        hI.setSimilarityScore(targetHit.similarity)
                    elif searchType == "substructure":
                        hI.setMatchedAtomLength(len(targetHit.match_atoms()))
                except Exception as e:
                    logger.exception("Failing with %s", str(e))
                    #
                #
                mU.mkdir(dirPath)
                mol2L = []
                for jj, mc in enumerate(targetHit.molecule.components, 1):
                    fp = os.path.join(dirPath, queryTargetId + "_" + targetHit.identifier + "_%03d" % jj + ".mol2")
                    mol2L.append(fp)
                    with MoleculeWriter(fp) as ofh:
                        ofh.write(mc)
                    # Replace the title line
                    with open(fp) as fin:
                        lines = fin.readlines()
                    lines[1] = lines[1].replace("00", targetHit.identifier)
                    #
                    with open(fp, "w") as fout:
                        fout.write("".join(lines))
                    #
                    fp = os.path.join(dirPath, queryTargetId + "_" + targetHit.identifier + "_%03d" % jj + ".sdf")
                    with MoleculeWriter(fp) as ofh:
                        ofh.write(mc)
                    # Replace the title line
                    with open(fp) as fin:
                        lines = fin.readlines()
                    lines[0] = lines[0].replace("00", targetHit.identifier)
                    #
                    with open(fp, "w") as fout:
                        fout.write("".join(lines))
                #
                #  Check for multiple generated result files -
                #
                for jj, fp in enumerate(mol2L, 1):
                    logger.debug("(%d) adding component fp %s", jj, fp)
                    hI.setMatchNumber(jj)
                    hI.setMol2Path(fp)
                    tt = fp[:-4] + "sdf"
                    hI.setMolPath(tt)
                    summaryList.append(copy.deepcopy(hI.get()))
                    #
        else:
            logger.info("(%d) se sarch for %s returns no matches", ii, queryTargetId)
            hits = None
        #
        if numHits > 0:
            mU.mkdir(dirPath)
            fp = os.path.join(dirPath, queryTargetId + "-index.json")
            cmI = CcdcMatchIndex(indexFilePath=fp, verbose=self.__verbose)
            cmI.load(summaryList)
            cmI.writeIndex()

        return numHits

    def __moleculeSubstructureSearch(self, aMol, suppressMetals=False):
        ms = MoleculeSubstructure(aMol)
        search = SubstructureSearch()
        search.add_substructure(ms)
        search.settings.has_3d_coordinates = True
        search.settings.no_disorder = True
        if suppressMetals:
            search.settings.only_organic = True
            search.settings.no_metals = True
        search.settings.max_r_factor = self.__rValueMaxPercent
        hits = search.search(max_hits_per_structure=1)
        return hits

    def __smartsSubstructureSearch(self, smarts, suppressMetals=False):
        ss = SMARTSSubstructure(smarts)
        search = SubstructureSearch()
        search.add_substructure(ss)
        search.settings.has_3d_coordinates = True
        search.settings.no_disorder = True
        if suppressMetals:
            search.settings.no_metals = True
            search.settings.only_organic = True
        search.settings.max_r_factor = self.__rValueMaxPercent
        hits = search.search(max_hits_per_structure=1)
        return hits

    def __similaritySearch(self, aMol, suppressMetals=False):
        # the similarity threshold is a score from 0 to 1 of how 'similar' the structures will be to the input molecule
        search = SimilaritySearch(aMol, threshold=self.__similarityThreshold)
        search.settings.has_3d_coordinates = True
        search.settings.no_disorder = True
        if suppressMetals:
            search.settings.only_organic = True
            search.settings.no_metals = True
        search.settings.max_r_factor = self.__rValueMaxPercent
        hits = search.search(max_hits_per_structure=1)
        return hits

    def __textSearch(self, aMol):
        search = TextNumericSearch()
        search.add_compound_name(aMol.identifier, mode="anywhere", ignore_non_alpha_num=True)
        # search.add_synonym(m.identifier) - no need to add synonym separate (should automatically search both name and synonym)
        textHits = search.search(max_hits_per_structure=1)
        return textHits



##
#
# File:    CcdcSearchMp.py
# Author:  J. Westbrook
# Date:    15-Dec-2020
# Version: 0.001
#
# Updated:
#
##
"""
MP execution wrapper for substructure and similarity search against the CCDC local Python API -

NOTE:  The latest ccdc API appears not compatible with rcsb.utils.multiproc and Python multiprocessing modules
       so this code is not currently used.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

# pylint: disable=redefined-outer-name

import logging
import time
import os
import os.path

from rcsb.utils.ccdc.CcdcSearch import CcdcSearch
from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil

logger = logging.getLogger(__name__)


class CcdcSearchWorker(object):
    def __init__(self, verbose=True):
        self.__verbose = verbose

    def __checkStop(self, path):
        try:
            if os.access(path, os.F_OK):
                return True
        except Exception:
            pass
        return False

    def search(self, dataList, procName, optionsD, workingDir):
        """"""
        resultPath = optionsD["resultPath"]
        searchType = optionsD["searchType"]
        _ = workingDir
        ky = None
        startTime = time.time()
        logger.info("starting %s at %s", procName, time.strftime("%Y %m %d %H:%M:%S", time.localtime()))
        resultList = []
        try:
            stopPath = os.path.join(resultPath, "STOP")
            logger.info("%s search list length %d", procName, len(dataList))

            for ii, queryTargetPath in enumerate(dataList):
                _, fn = os.path.split(queryTargetPath)
                queryTargetId, _ = os.path.splitext(fn)
                parentId = queryTargetId.split("|")[0]
                logger.info("%s starting %s (%s) (%d of %d)", procName, parentId, queryTargetId, ii + 1, len(dataList))
                if self.__checkStop(stopPath):
                    logger.info("%s stopping at %d of %d", procName, ii + 1, len(dataList))
                    break
                logger.info("%s search list %d of %d", procName, ii + 1, len(dataList))
                try:
                    vS = CcdcSearch(verbose=self.__verbose)
                    numHits = vS.search(queryTargetId, queryTargetPath, resultPath, searchType=searchType)
                    if numHits:
                        resultList.append(queryTargetPath)
                except Exception as e:
                    logger.exception("Failing for %r with %s", ky, str(e))

        except Exception as e:
            logger.exception("Failing for %r with %s", ky, str(e))

        endTime = time.time()
        logger.info(" %s completed at %s (%.2f seconds)", procName, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
        return resultList, [], []


class CcdcSearchMp(object):
    def __init__(self, verbose=True):
        """MP execution wrapper for variant search against the CCDC local Python API -"""
        self.__verbose = verbose
        #

    def runSearch(self, molFilePathList, resultPath, searchType="similarity", numProc=4, chunkSize=10):
        """"""
        logger.info("Starting with molfile path list length %d", len(molFilePathList))
        try:
            pU = CcdcSearchWorker(verbose=self.__verbose)
            mpu = MultiProcUtil(verbose=True)
            # mpu.setWorkingDir(resultPath)
            mpu.setOptions(optionsD={"resultPath": resultPath, "searchType": searchType})
            #
            mpu.set(workerObj=pU, workerMethod="search")

            ok, failList, resultList, _ = mpu.runMulti(dataList=molFilePathList, numProc=numProc, numResults=1, chunkSize=chunkSize)
            logger.info("run ended status %r success count %d failures %r", ok, len(resultList[0]), len(failList))
        except Exception as e:
            logger.exception("Failing with %s", str(e))
